from logging import getLogger
from collections import defaultdict
from random import shuffle, choice
import libtcodpy as tcod

log = getLogger("rainbowrooms.generator")

class GenerationError(Exception):
    pass

class GeneratorState(object):
    def __init__(self):
        self.start = None
        self.ramps = set()

def generate(width, height, bottom, state):
    # Use the switch colors here; walls are fixed in wall().
    colors = "rygbv"
    
    if not state:
        state = GeneratorState()
    newstate = GeneratorState()
    
    def generate_walls():
        cells = dict(((x,y),       (x,y)) for x in range(width)   for y in range(height))
        walls = dict(((x,y,True),  None)  for x in range(width-1) for y in range(height))
        walls.update(((x,y,False), None)  for x in range(width)   for y in range(height-1))
        keys = list(walls.keys())
        shuffle(keys)
        
        for key in state.ramps:
            # Mark any ramps back up to the previous level.
            walls[key] = "<"
        
        def root(cell):
            parent = cells[cell]
            if parent != cell:
                cells[cell] = parent = root(parent)
            return parent
        
        for key in keys:
            x, y, vertical = key
            if vertical:
                other = (x+1,y)
            else:
                other = (x,y+1)
            
            if state.start in [(x,y), other]:
                # Don't connect the cell behind the ramp.
                continue
            
            parent = root(other)
            main = root((x,y))
            if parent != main:
                walls[key] = choice(colors + "+++++")
                cells[parent] = main
        
        return walls

    def directions(x,y):
        if x:
            yield (x-1, y, True),  (x-1, y)
        if y:
            yield (x, y-1, False), (x, y-1)
        if x < width-1:
            yield (x, y, True),    (x+1, y)
        if y < height-1:
            yield (x, y, False),   (x, y+1)

    def generate_ends():
        cells = dict(((x,y), "+") for x in range(width) for y in range(height))
        cell = state.start or (0,0)
        
        # Find a goal cell
        last = expand(cell, [])[1]
        goal = choice(last)
        if bottom:
            cells[goal] = "$"
        else:
            cells[goal] = " "
            for wall, other in directions(*goal):
                if walls[wall]:
                    newstate.start = other
                    newstate.ramps.add(wall)
                    walls[wall] = ">"
        
        if state.start:
            start = state.start
            cells[start] = " "
        else:
            # Find a start cell
            last = expand(goal, [])[1]
            start = choice(last)
            cells[start] = "@"
            
            # Don't let the start be blocked by a door.
            for wall, other in directions(*start):
                if walls[wall]:
                    walls[wall] = "+"
        
        log.debug("Ends: %s/%s -> %s", cell, start, goal)
        return cells, start, goal
    
    def goalside(cell, wall):
        # Returns True if the goal can be reached without going through the
        # door from the cell, but the start cannot.
        seen = set()
        boundary = [cell]
        while boundary:
            if start in boundary:
                return False
            seen.update(boundary)
            boundary = [
                c
                for neighbor in boundary
                for w,c in directions(*neighbor)
                if c not in seen and w != wall and walls[w]
            ]
        return goal in seen
    
    def expand(cell, disallowed):
        allowed = set(colors + "<+>") - set(disallowed)
        seen = set()
        boundary = [cell]
        while boundary:
            last = boundary
            seen.update(boundary)
            boundary = [
                c
                for neighbor in boundary
                for w,c in directions(*neighbor)
                if c not in seen and walls[w] in allowed
            ]
        return seen, last
    
    def enemies(color):
        c = colors.find(color)
        return set(colors[(c+i)%5] for i in (2,3))
    
    def accessible(wall, side):
        # A switch is accessible from here if we can reach it:
        # 1) without going through any doors;
        # 2) by going through doors of exactly one ally, whose switch is otherwise accessible; or
        # 3) by going through doors of both allies, where the first door has a switch between the two,
        #    and the second door has a switch accessible from the other side of the first door.
        # Todo: Implement option 3?
        blocked = set(colors)
        area, last = expand(side, blocked)
        switches = set(cells[c] for c in area)
        color = walls[wall]
        c = colors.find(color)
        cycle = (colors*2)[c+1:c+5]
        for path in cycle, list(reversed(cycle)):
            # A cycle of colors can be followed, but only claim access to
            # spaces where a switch of the wall's color would be accessible.
            for n in range(3):
                if path[n] in switches:
                    allowed = True
                    break
                expanded = expand(side, blocked - set(path[n+1]))[0]
                if not any(cells[loc] == path[n] for loc in expanded):
                    allowed = False
                    break
            else:
                allowed = path[3] in switches
            
            if allowed:
                ally = path[0]
                behind, last = expand(side, blocked - set(ally))
                area.update(behind)
        return area, last
    
    def generate_switches():
        placed = set()
        sides = {}
        for wall in walls:
            color = walls[wall]
            if color and color in colors:
                x, y, vertical = wall
                if vertical:
                    sides[wall] = (x,y), (x+1,y)
                else:
                    sides[wall] = (x,y), (x,y+1)
        
        while sides:
            log.debug("Checking %d walls", len(sides))
            newsides = defaultdict(set)
            for wall in sides:
                added = None
                color = walls[wall]
                for cell in sides[wall]:
                    pos = place_switch(cell, wall)
                    if pos:
                        placed.add(pos)
                    elif walls[wall] != color:
                        # Door has been removed.
                        if added:
                            # Remove its other switch.
                            # This may only be done before looking at other doors.
                            cells[added] = "+"
                        
                        # Re-run any neighboring doors; they might not be isolated anymore.
                        for cell in expand(cell, colors)[0]:
                            for wall, other in directions(*cell):
                                color = walls[wall]
                                if color and color in colors:
                                    newsides[wall].add(cell)
                        
                        # Don't bother with the other side.
                        break
                    added = pos
            sides = newsides
        
        return placed
    
    def place_switch(cell, wall):
        # Direction of the wall relative to the cell.
        # Only useful for the debugging output.
        if cell[0] > wall[0]:
            direction = "W"
        elif cell[1] > wall[1]:
            direction = "N"
        elif wall[2]:
            direction = "E"
        else:
            direction = "S"
        #log.debug("%s %s: %s", cell, direction, wall)
        
        color = walls[wall]
        if color not in colors:
            log.debug("%s %s: Colorless", cell, direction)
            return
        
        if goalside(cell, wall):
            # No need to go back.
            log.debug("%s %s: Goal side", cell, direction)
            return
        
        possible, furthest = accessible(wall, cell)
        switches = set(cells[c] for c in possible)
        if color in switches:
            # An accessible switch already exists.
            log.debug("%s %s: Served by %r", cell, direction, [c for c in possible if cells[c] == color])
            return
        
        area = expand(cell, color)[0]
        if start not in area and area == possible:
            log.debug("%s %s: Isolated", cell, direction)
            return
        
        empty = [c for c in furthest if cells[c] == "+"]
        if not empty:
            # Can't be placed as far away as possible,
            # so put it anywhere we can reach.
            empty = [c for c in possible if cells[c] == "+"]
        
        if not empty:
            # Unable to find space for the switch.
            # Remove the problematic wall.
            log.debug("%s %s: No space available", cell, direction)
            walls[wall] = "+"
            return
        
        pos = choice(empty)
        cells[pos] = color
        log.debug("%s %s: Adding %s", cell, direction, pos)
        return pos
    
    def trim_switches(trim):
        # Trim any useless switches
        # Switches that still don't need to exist:
        # - @++bgGbB+$: A switch on the goal side of an allied door,
        #   with both switches on the start side of that door,
        #   and no way to turn off either one in between.
        while trim:
            log.debug("Checking %s switches...", len(trim))
            trimmed = False
            untrimmed = set()
            for cell in trim:
                switch = cells[cell]
                if switch not in colors:
                    # Already cleared.
                    continue
                
                influence = expand(cell, switch)[0]
                if start not in influence:
                    disabling = enemies(switch)
                    if not any(cells[c] in disabling for c in influence):
                        # There's no other way to turn this color off.
                        log.debug("Trimming %s %s: Unnecessary", cell, switch)
                        cells[cell] = "+"
                        trimmed = True
                        continue
                
                doors = all(
                    any(cells[c] == switch and c != cell for c in accessible(door,side)[0])
                    for side in influence
                    for door,other in directions(*side)
                    if walls[door] == switch
                )
                
                if doors:
                    # Each of the served doors is also served by another.
                    log.debug("Trimming %s %s: Redundant", cell, switch)
                    cells[cell] = "+"
                    trimmed = True
                    continue
                
                area = expand(cell, colors)[0]
                if any(cells[c] == switch and c != cell for c in area):
                    log.debug("Trimming %s %s: Duplicate", cell, switch)
                    cells[cell] = "+"
                    trimmed = True
                    continue
                
                untrimmed.add(cell)
            
            if trimmed:
                trim = untrimmed
            else:
                break
    
    while True:
        try:
            walls = generate_walls()
            cells, start, goal = generate_ends()
            switches = generate_switches()
            trim_switches(switches)
        except GenerationError:
            log.exception("Level generation failed; retrying...")
        else:
            break
    
    for line in map_lines(cells, walls, width, height):
        log.debug(str.join("", line))
    
    return colors, cells, walls, newstate

def map_lines(cells, walls, width, height):
    def wall(value):
        return (value or " ").upper()
    
    for y in range(height-1):
        line = []
        for x in range(width-1):
            line.append(cells[x,y])
            line.append(wall(walls[x,y,True]))
        line.append(cells[width-1,y])
        yield line
        
        line = []
        for x in range(width):
            line.append(wall(walls[x,y,False]))
            line.append(" ")
        yield line
    
    y = height - 1
    line = []
    for x in range(width-1):
        line.append(cells[x,y])
        line.append(wall(walls[x,y,True]))
    line.append(cells[width-1,y])
    yield line

def display():
    w, h = 40, 20
    for line in generate(w, h):
        print(str.join("", line))
    
    tcod.console_init_root(w*3, h*3, "Map Generation", False)
    color = tcod.grey
    for y, line in enumerate(generate(w, h)):
        for x, char in enumerate(line):
            if char == "+":
                tcod.console_set_char_background(0, x+2,y+2, color)
    tcod.console_flush()
    tcod.console_wait_for_keypress(True)

if __name__ == "__main__":
    display()
