from collections import defaultdict
from rainbowrooms.fov import fieldOfView
from rainbowrooms.world import Rock, Tile, Wall
import libtcodpy as tcod
import logging

class Window(object):
    def __init__(self):
        # Calculate screen sizes, for a maximum maze size of 25x17.
        self.column = 20
        mapwidth = 25*4 + 1
        mapheight = 17*4 + 1
        self.width = mapwidth + self.column + 2
        self.height = mapheight + 4
        
        # Basic libtcod initialization
        tcod.console_init_root(self.width, self.height, b"Rainbow Rooms", False)
        
        # Logging
        logging.basicConfig(level=logging.INFO)
        self.logger = logging.getLogger("rainbowrooms")
        
        # Subwindows
        self.map = tcod.console_new(mapwidth, mapheight)
        self.msg = None
        
        # Key names
        self.keynames = {}
        for name in dir(tcod):
            if name.startswith("KEY_") and name not in ("KEY_PRESSED", "KEY_RELEASED"):
                self.keynames[getattr(tcod, name)] = name[4:]
    
    def display(self, world):
        tcod.console_clear(self.map)
        px, py, pz = world.player
        
        h = tcod.console_get_height(self.map)
        w = tcod.console_get_width(self.map)
        assert 0 <= px < w
        assert 0 <= py < h
        radius = h+w
        ploc = (px, py)
        
        seen = {}
        neighbors = defaultdict(set)
        neighbors[ploc] = set(pz)
        spread = [[-1,0], [-1,0,1], [0,1]]
        
        def visit(x, y):
            z = neighbors[x,y]
            #self.log("Visiting %s,%s: %s", x, y, z)
            assert z
            tile = world[x, y, z]
            seen[x,y] = tile
            if tile.transparent:
                if tile.passable:
                    z = tile.passable
                for dx in spread[cmp(x, px) + 1]:
                    for dy in spread[cmp(y, py) + 1]:
                        loc = x+dx, y+dy
                        neighbors[loc].update(z)
                        #self.log("Updating %s to %s", loc, neighbors[loc])
        
        def blocked(x, y):
            #self.log("Checking %s,%s", x, y)
            return not seen[x,y].transparent
        
        fieldOfView(px, py, w, h, radius, visit, blocked)
        
        for loc in seen:
            space = seen[loc]
            self.cell(loc, space.symbol, space.color, Tile.color)
        self.cell(ploc, tcod.CHAR_SMILIE, tcod.lightest_grey)
        
        tcod.console_blit(self.map, 0,0, 0,0, 0, self.column,2)
        
        # Maze number
        tcod.console_print_ex(0, self.column>>1, 2, tcod.BKGND_NONE, tcod.CENTER, " Level " + str(world.maze))
        
        # Guide to each color's state.
        offset = (self.column - 12) >> 1
        locs = [(0,2), (6,0), (10,5), (6,10), (0,8)]
        pixels = [
            [0x80, 0x20, 0x90],
            [0x20, 0x20, 0x20],
            [0x81, 0x20, 0x91],
        ]
        
        for n, (x,y) in enumerate(locs):
            color = Wall.colors[n][Wall.switched[n]]
            for dy, row in enumerate(pixels):
                for dx, char in enumerate(row):
                    tcod.console_put_char_ex(0, x+dx+offset, y+dy+5, char, tcod.black, color)
        
        if self.msg:
            self.display_message()
        
        tcod.console_flush()
    
    def display_message(self):
        textwidth = (self.width >> 1) - 4
        textheight = tcod.console_get_height_rect(0, 0, 0, textwidth, self.height - 4, self.msg)
        if textheight == 5:
            # The message isn't wrapped; crop the box to its width.
            textwidth = len(self.msg)
        elif textheight < self.height - 4:
            # The message box might not need to be quite so wide.
            while textwidth > 1:
                height = tcod.console_get_height_rect(0, 0, 0, textwidth - 1, self.height, self.msg)
                if height > textheight:
                    break
                textwidth -= 1
        else:
            # The message box needs to be wider to accommodate the height.
            while textwidth < self.width - 4:
                textwidth += 1
                textheight = tcod.console_get_height_rect(0, 0, 0, textwidth, textheight, self.msg)
                if textheight < self.height - 4:
                    break
        width = textwidth + 4
        height = textheight + 4
        window = tcod.console_new(width, height)
        tcod.console_set_default_foreground(window, tcod.white)
        tcod.console_print_frame(window, 0, 0, width, height, False, tcod.BKGND_NONE)
        tcod.console_print_rect_ex(window, width >> 1, 2, textwidth, height, tcod.BKGND_NONE, tcod.CENTER, self.msg)
        
        #blit the contents of "window" to the root console
        x = (self.width>>1) - (width>>1)
        y = (self.height>>1) - (height>>1)
        tcod.console_blit(window, 0, 0, width, height, 0, x, y, 1.0, 0.7)
    
    def cell(self, loc, symbol, color=None, bg=None):
        try:
            # Map strings to numbers.
            symbol = ord(symbol)
        except TypeError:
            pass
        
        if not symbol:
            symbol = " "
            bg, color = color, bg
        
        x, y = loc
        assert 0 <= x < tcod.console_get_width(self.map)
        assert 0 <= y < tcod.console_get_height(self.map)
        if color: tcod.console_set_char_foreground(self.map, x,y, color)
        if bg: tcod.console_set_char_background(self.map, x,y, bg)
        tcod.console_set_char(self.map, x,y, symbol)
    
    def event(self):
        self.msg = None
        char = tcod.Key()
        mouse = tcod.Mouse()
        pressed = None
        
        # Ignore key up and mouse movement events.
        # Accepting key down allows us to use repeats,
        # but accepting both up and down duplicates commands.
        eventmask = tcod.EVENT_KEY_PRESS | tcod.EVENT_MOUSE_PRESS | tcod.EVENT_MOUSE_RELEASE
        
        while not tcod.console_is_window_closed():
            try:
                tcod.sys_wait_for_event(eventmask, char, mouse, False)
                
                if char.vk == tcod.KEY_NONE:
                    # Not a key event; probably a mouse event instead.
                    if mouse.lbutton:
                        pressed = mouse.cx, mouse.cy
                    elif mouse.lbutton_pressed and mouse.cx and mouse.cy and (mouse.cx, mouse.cy) == pressed:
                        return pressed
                    continue
                
                if char.vk == tcod.KEY_CHAR:
                    key = chr(char.c)
                else:
                    key = self.keynames.get(char.vk, str(char.vk))
                    if char.shift:
                        key = "S-" + key
                
                if char.lctrl or char.rctrl:
                    key = "^" + key
            except KeyboardInterrupt:
                key = "^c"
            
            return key
        return "Quit"
    
    def message(self, msg):
        self.msg = msg
    
    def log(self, message, *args):
        if args:
            format = message
        else:
            format = "%s"
            args = message,
        self.logger.info(format, *args)
